/****************************************************************************
** Meta object code from reading C++ file 'XpdfViewer.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.9.4)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../src/xpdf-4.03/xpdf-qt/XpdfViewer.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'XpdfViewer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.9.4. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_XpdfMenuButton_t {
    QByteArrayData data[3];
    char stringdata0[27];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_XpdfMenuButton_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_XpdfMenuButton_t qt_meta_stringdata_XpdfMenuButton = {
    {
QT_MOC_LITERAL(0, 0, 14), // "XpdfMenuButton"
QT_MOC_LITERAL(1, 15, 10), // "btnPressed"
QT_MOC_LITERAL(2, 26, 0) // ""

    },
    "XpdfMenuButton\0btnPressed\0"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_XpdfMenuButton[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
       1,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,   19,    2, 0x08 /* Private */,

 // slots: parameters
    QMetaType::Void,

       0        // eod
};

void XpdfMenuButton::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        XpdfMenuButton *_t = static_cast<XpdfMenuButton *>(_o);
        Q_UNUSED(_t)
        switch (_id) {
        case 0: _t->btnPressed(); break;
        default: ;
        }
    }
    Q_UNUSED(_a);
}

const QMetaObject XpdfMenuButton::staticMetaObject = {
    { &QToolButton::staticMetaObject, qt_meta_stringdata_XpdfMenuButton.data,
      qt_meta_data_XpdfMenuButton,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *XpdfMenuButton::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *XpdfMenuButton::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_XpdfMenuButton.stringdata0))
        return static_cast<void*>(this);
    return QToolButton::qt_metacast(_clname);
}

int XpdfMenuButton::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QToolButton::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 1;
    }
    return _id;
}
struct qt_meta_stringdata_XpdfErrorWindow_t {
    QByteArrayData data[3];
    char stringdata0[33];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_XpdfErrorWindow_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_XpdfErrorWindow_t qt_meta_stringdata_XpdfErrorWindow = {
    {
QT_MOC_LITERAL(0, 0, 15), // "XpdfErrorWindow"
QT_MOC_LITERAL(1, 16, 15), // "clearBtnPressed"
QT_MOC_LITERAL(2, 32, 0) // ""

    },
    "XpdfErrorWindow\0clearBtnPressed\0"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_XpdfErrorWindow[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
       1,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,   19,    2, 0x08 /* Private */,

 // slots: parameters
    QMetaType::Void,

       0        // eod
};

void XpdfErrorWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        XpdfErrorWindow *_t = static_cast<XpdfErrorWindow *>(_o);
        Q_UNUSED(_t)
        switch (_id) {
        case 0: _t->clearBtnPressed(); break;
        default: ;
        }
    }
    Q_UNUSED(_a);
}

const QMetaObject XpdfErrorWindow::staticMetaObject = {
    { &QWidget::staticMetaObject, qt_meta_stringdata_XpdfErrorWindow.data,
      qt_meta_data_XpdfErrorWindow,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *XpdfErrorWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *XpdfErrorWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_XpdfErrorWindow.stringdata0))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int XpdfErrorWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 1;
    }
    return _id;
}
struct qt_meta_stringdata_XpdfViewer_t {
    QByteArrayData data[82];
    char stringdata0[1400];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_XpdfViewer_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_XpdfViewer_t qt_meta_stringdata_XpdfViewer = {
    {
QT_MOC_LITERAL(0, 0, 10), // "XpdfViewer"
QT_MOC_LITERAL(1, 11, 22), // "remoteServerConnection"
QT_MOC_LITERAL(2, 34, 0), // ""
QT_MOC_LITERAL(3, 35, 16), // "remoteServerRead"
QT_MOC_LITERAL(4, 52, 10), // "pdfResized"
QT_MOC_LITERAL(5, 63, 12), // "pdfPaintDone"
QT_MOC_LITERAL(6, 76, 8), // "finished"
QT_MOC_LITERAL(7, 85, 7), // "preLoad"
QT_MOC_LITERAL(8, 93, 8), // "postLoad"
QT_MOC_LITERAL(9, 102, 8), // "keyPress"
QT_MOC_LITERAL(10, 111, 10), // "QKeyEvent*"
QT_MOC_LITERAL(11, 122, 1), // "e"
QT_MOC_LITERAL(12, 124, 10), // "mousePress"
QT_MOC_LITERAL(13, 135, 12), // "QMouseEvent*"
QT_MOC_LITERAL(14, 148, 12), // "mouseRelease"
QT_MOC_LITERAL(15, 161, 10), // "mouseClick"
QT_MOC_LITERAL(16, 172, 16), // "mouseDoubleClick"
QT_MOC_LITERAL(17, 189, 16), // "mouseTripleClick"
QT_MOC_LITERAL(18, 206, 10), // "mouseWheel"
QT_MOC_LITERAL(19, 217, 12), // "QWheelEvent*"
QT_MOC_LITERAL(20, 230, 9), // "mouseMove"
QT_MOC_LITERAL(21, 240, 10), // "pageChange"
QT_MOC_LITERAL(22, 251, 2), // "pg"
QT_MOC_LITERAL(23, 254, 20), // "sidebarSplitterMoved"
QT_MOC_LITERAL(24, 275, 3), // "pos"
QT_MOC_LITERAL(25, 279, 5), // "index"
QT_MOC_LITERAL(26, 285, 14), // "openMenuAction"
QT_MOC_LITERAL(27, 300, 22), // "openInNewWinMenuAction"
QT_MOC_LITERAL(28, 323, 16), // "reloadMenuAction"
QT_MOC_LITERAL(29, 340, 16), // "saveAsMenuAction"
QT_MOC_LITERAL(30, 357, 19), // "saveImageMenuAction"
QT_MOC_LITERAL(31, 377, 14), // "quitMenuAction"
QT_MOC_LITERAL(32, 392, 14), // "copyMenuAction"
QT_MOC_LITERAL(33, 407, 24), // "singlePageModeMenuAction"
QT_MOC_LITERAL(34, 432, 24), // "continuousModeMenuAction"
QT_MOC_LITERAL(35, 457, 30), // "sideBySideSingleModeMenuAction"
QT_MOC_LITERAL(36, 488, 34), // "sideBySideContinuousModeMenuA..."
QT_MOC_LITERAL(37, 523, 34), // "horizontalContinuousModeMenuA..."
QT_MOC_LITERAL(38, 558, 20), // "fullScreenMenuAction"
QT_MOC_LITERAL(39, 579, 7), // "checked"
QT_MOC_LITERAL(40, 587, 25), // "rotateClockwiseMenuAction"
QT_MOC_LITERAL(41, 613, 32), // "rotateCounterclockwiseMenuAction"
QT_MOC_LITERAL(42, 646, 25), // "zoomToSelectionMenuAction"
QT_MOC_LITERAL(43, 672, 23), // "toggleToolbarMenuAction"
QT_MOC_LITERAL(44, 696, 23), // "toggleSidebarMenuAction"
QT_MOC_LITERAL(45, 720, 24), // "viewPageLabelsMenuAction"
QT_MOC_LITERAL(46, 745, 16), // "newTabMenuAction"
QT_MOC_LITERAL(47, 762, 19), // "newWindowMenuAction"
QT_MOC_LITERAL(48, 782, 18), // "closeTabMenuAction"
QT_MOC_LITERAL(49, 801, 21), // "closeWindowMenuAction"
QT_MOC_LITERAL(50, 823, 25), // "openErrorWindowMenuAction"
QT_MOC_LITERAL(51, 849, 14), // "helpMenuAction"
QT_MOC_LITERAL(52, 864, 21), // "keyBindingsMenuAction"
QT_MOC_LITERAL(53, 886, 15), // "aboutMenuAction"
QT_MOC_LITERAL(54, 902, 15), // "popupMenuAction"
QT_MOC_LITERAL(55, 918, 3), // "idx"
QT_MOC_LITERAL(56, 922, 26), // "toggleSidebarButtonPressed"
QT_MOC_LITERAL(57, 949, 17), // "pageNumberChanged"
QT_MOC_LITERAL(58, 967, 17), // "backButtonPressed"
QT_MOC_LITERAL(59, 985, 20), // "forwardButtonPressed"
QT_MOC_LITERAL(60, 1006, 20), // "zoomOutButtonPressed"
QT_MOC_LITERAL(61, 1027, 19), // "zoomInButtonPressed"
QT_MOC_LITERAL(62, 1047, 16), // "zoomIndexChanged"
QT_MOC_LITERAL(63, 1064, 8), // "zoomText"
QT_MOC_LITERAL(64, 1073, 19), // "zoomEditingFinished"
QT_MOC_LITERAL(65, 1093, 21), // "fitWidthButtonPressed"
QT_MOC_LITERAL(66, 1115, 20), // "fitPageButtonPressed"
QT_MOC_LITERAL(67, 1136, 23), // "selectModeButtonPressed"
QT_MOC_LITERAL(68, 1160, 22), // "statusIndicatorPressed"
QT_MOC_LITERAL(69, 1183, 15), // "findTextChanged"
QT_MOC_LITERAL(70, 1199, 21), // "findNextButtonPressed"
QT_MOC_LITERAL(71, 1221, 21), // "findPrevButtonPressed"
QT_MOC_LITERAL(72, 1243, 19), // "newTabButtonPressed"
QT_MOC_LITERAL(73, 1263, 9), // "switchTab"
QT_MOC_LITERAL(74, 1273, 16), // "QListWidgetItem*"
QT_MOC_LITERAL(75, 1290, 7), // "current"
QT_MOC_LITERAL(76, 1298, 8), // "previous"
QT_MOC_LITERAL(77, 1307, 19), // "infoComboBoxChanged"
QT_MOC_LITERAL(78, 1327, 18), // "outlineItemClicked"
QT_MOC_LITERAL(79, 1346, 16), // "layerItemClicked"
QT_MOC_LITERAL(80, 1363, 21), // "attachmentSaveClicked"
QT_MOC_LITERAL(81, 1385, 14) // "clearFindError"

    },
    "XpdfViewer\0remoteServerConnection\0\0"
    "remoteServerRead\0pdfResized\0pdfPaintDone\0"
    "finished\0preLoad\0postLoad\0keyPress\0"
    "QKeyEvent*\0e\0mousePress\0QMouseEvent*\0"
    "mouseRelease\0mouseClick\0mouseDoubleClick\0"
    "mouseTripleClick\0mouseWheel\0QWheelEvent*\0"
    "mouseMove\0pageChange\0pg\0sidebarSplitterMoved\0"
    "pos\0index\0openMenuAction\0"
    "openInNewWinMenuAction\0reloadMenuAction\0"
    "saveAsMenuAction\0saveImageMenuAction\0"
    "quitMenuAction\0copyMenuAction\0"
    "singlePageModeMenuAction\0"
    "continuousModeMenuAction\0"
    "sideBySideSingleModeMenuAction\0"
    "sideBySideContinuousModeMenuAction\0"
    "horizontalContinuousModeMenuAction\0"
    "fullScreenMenuAction\0checked\0"
    "rotateClockwiseMenuAction\0"
    "rotateCounterclockwiseMenuAction\0"
    "zoomToSelectionMenuAction\0"
    "toggleToolbarMenuAction\0toggleSidebarMenuAction\0"
    "viewPageLabelsMenuAction\0newTabMenuAction\0"
    "newWindowMenuAction\0closeTabMenuAction\0"
    "closeWindowMenuAction\0openErrorWindowMenuAction\0"
    "helpMenuAction\0keyBindingsMenuAction\0"
    "aboutMenuAction\0popupMenuAction\0idx\0"
    "toggleSidebarButtonPressed\0pageNumberChanged\0"
    "backButtonPressed\0forwardButtonPressed\0"
    "zoomOutButtonPressed\0zoomInButtonPressed\0"
    "zoomIndexChanged\0zoomText\0zoomEditingFinished\0"
    "fitWidthButtonPressed\0fitPageButtonPressed\0"
    "selectModeButtonPressed\0statusIndicatorPressed\0"
    "findTextChanged\0findNextButtonPressed\0"
    "findPrevButtonPressed\0newTabButtonPressed\0"
    "switchTab\0QListWidgetItem*\0current\0"
    "previous\0infoComboBoxChanged\0"
    "outlineItemClicked\0layerItemClicked\0"
    "attachmentSaveClicked\0clearFindError"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_XpdfViewer[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
      66,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,  344,    2, 0x08 /* Private */,
       3,    0,  345,    2, 0x08 /* Private */,
       4,    0,  346,    2, 0x08 /* Private */,
       5,    1,  347,    2, 0x08 /* Private */,
       7,    0,  350,    2, 0x08 /* Private */,
       8,    0,  351,    2, 0x08 /* Private */,
       9,    1,  352,    2, 0x08 /* Private */,
      12,    1,  355,    2, 0x08 /* Private */,
      14,    1,  358,    2, 0x08 /* Private */,
      15,    1,  361,    2, 0x08 /* Private */,
      16,    1,  364,    2, 0x08 /* Private */,
      17,    1,  367,    2, 0x08 /* Private */,
      18,    1,  370,    2, 0x08 /* Private */,
      20,    1,  373,    2, 0x08 /* Private */,
      21,    1,  376,    2, 0x08 /* Private */,
      23,    2,  379,    2, 0x08 /* Private */,
      26,    0,  384,    2, 0x08 /* Private */,
      27,    0,  385,    2, 0x08 /* Private */,
      28,    0,  386,    2, 0x08 /* Private */,
      29,    0,  387,    2, 0x08 /* Private */,
      30,    0,  388,    2, 0x08 /* Private */,
      31,    0,  389,    2, 0x08 /* Private */,
      32,    0,  390,    2, 0x08 /* Private */,
      33,    0,  391,    2, 0x08 /* Private */,
      34,    0,  392,    2, 0x08 /* Private */,
      35,    0,  393,    2, 0x08 /* Private */,
      36,    0,  394,    2, 0x08 /* Private */,
      37,    0,  395,    2, 0x08 /* Private */,
      38,    1,  396,    2, 0x08 /* Private */,
      40,    0,  399,    2, 0x08 /* Private */,
      41,    0,  400,    2, 0x08 /* Private */,
      42,    0,  401,    2, 0x08 /* Private */,
      43,    1,  402,    2, 0x08 /* Private */,
      44,    1,  405,    2, 0x08 /* Private */,
      45,    1,  408,    2, 0x08 /* Private */,
      46,    0,  411,    2, 0x08 /* Private */,
      47,    0,  412,    2, 0x08 /* Private */,
      48,    0,  413,    2, 0x08 /* Private */,
      49,    0,  414,    2, 0x08 /* Private */,
      50,    0,  415,    2, 0x08 /* Private */,
      51,    0,  416,    2, 0x08 /* Private */,
      52,    0,  417,    2, 0x08 /* Private */,
      53,    0,  418,    2, 0x08 /* Private */,
      54,    1,  419,    2, 0x08 /* Private */,
      56,    0,  422,    2, 0x08 /* Private */,
      57,    0,  423,    2, 0x08 /* Private */,
      58,    0,  424,    2, 0x08 /* Private */,
      59,    0,  425,    2, 0x08 /* Private */,
      60,    0,  426,    2, 0x08 /* Private */,
      61,    0,  427,    2, 0x08 /* Private */,
      62,    1,  428,    2, 0x08 /* Private */,
      64,    0,  431,    2, 0x08 /* Private */,
      65,    0,  432,    2, 0x08 /* Private */,
      66,    0,  433,    2, 0x08 /* Private */,
      67,    0,  434,    2, 0x08 /* Private */,
      68,    0,  435,    2, 0x08 /* Private */,
      69,    0,  436,    2, 0x08 /* Private */,
      70,    0,  437,    2, 0x08 /* Private */,
      71,    0,  438,    2, 0x08 /* Private */,
      72,    0,  439,    2, 0x08 /* Private */,
      73,    2,  440,    2, 0x08 /* Private */,
      77,    1,  445,    2, 0x08 /* Private */,
      78,    1,  448,    2, 0x08 /* Private */,
      79,    1,  451,    2, 0x08 /* Private */,
      80,    1,  454,    2, 0x08 /* Private */,
      81,    0,  457,    2, 0x08 /* Private */,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,    6,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 10,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, 0x80000000 | 19,   11,
    QMetaType::Void, 0x80000000 | 13,   11,
    QMetaType::Void, QMetaType::Int,   22,
    QMetaType::Void, QMetaType::Int, QMetaType::Int,   24,   25,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,   39,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,   39,
    QMetaType::Void, QMetaType::Bool,   39,
    QMetaType::Void, QMetaType::Bool,   39,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Int,   55,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::QString,   63,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 74, 0x80000000 | 74,   75,   76,
    QMetaType::Void, QMetaType::Int,   55,
    QMetaType::Void, QMetaType::QModelIndex,   55,
    QMetaType::Void, QMetaType::QModelIndex,   55,
    QMetaType::Void, QMetaType::Int,   55,
    QMetaType::Void,

       0        // eod
};

void XpdfViewer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        XpdfViewer *_t = static_cast<XpdfViewer *>(_o);
        Q_UNUSED(_t)
        switch (_id) {
        case 0: _t->remoteServerConnection(); break;
        case 1: _t->remoteServerRead(); break;
        case 2: _t->pdfResized(); break;
        case 3: _t->pdfPaintDone((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 4: _t->preLoad(); break;
        case 5: _t->postLoad(); break;
        case 6: _t->keyPress((*reinterpret_cast< QKeyEvent*(*)>(_a[1]))); break;
        case 7: _t->mousePress((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 8: _t->mouseRelease((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 9: _t->mouseClick((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 10: _t->mouseDoubleClick((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 11: _t->mouseTripleClick((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 12: _t->mouseWheel((*reinterpret_cast< QWheelEvent*(*)>(_a[1]))); break;
        case 13: _t->mouseMove((*reinterpret_cast< QMouseEvent*(*)>(_a[1]))); break;
        case 14: _t->pageChange((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 15: _t->sidebarSplitterMoved((*reinterpret_cast< int(*)>(_a[1])),(*reinterpret_cast< int(*)>(_a[2]))); break;
        case 16: _t->openMenuAction(); break;
        case 17: _t->openInNewWinMenuAction(); break;
        case 18: _t->reloadMenuAction(); break;
        case 19: _t->saveAsMenuAction(); break;
        case 20: _t->saveImageMenuAction(); break;
        case 21: _t->quitMenuAction(); break;
        case 22: _t->copyMenuAction(); break;
        case 23: _t->singlePageModeMenuAction(); break;
        case 24: _t->continuousModeMenuAction(); break;
        case 25: _t->sideBySideSingleModeMenuAction(); break;
        case 26: _t->sideBySideContinuousModeMenuAction(); break;
        case 27: _t->horizontalContinuousModeMenuAction(); break;
        case 28: _t->fullScreenMenuAction((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 29: _t->rotateClockwiseMenuAction(); break;
        case 30: _t->rotateCounterclockwiseMenuAction(); break;
        case 31: _t->zoomToSelectionMenuAction(); break;
        case 32: _t->toggleToolbarMenuAction((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 33: _t->toggleSidebarMenuAction((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 34: _t->viewPageLabelsMenuAction((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 35: _t->newTabMenuAction(); break;
        case 36: _t->newWindowMenuAction(); break;
        case 37: _t->closeTabMenuAction(); break;
        case 38: _t->closeWindowMenuAction(); break;
        case 39: _t->openErrorWindowMenuAction(); break;
        case 40: _t->helpMenuAction(); break;
        case 41: _t->keyBindingsMenuAction(); break;
        case 42: _t->aboutMenuAction(); break;
        case 43: _t->popupMenuAction((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 44: _t->toggleSidebarButtonPressed(); break;
        case 45: _t->pageNumberChanged(); break;
        case 46: _t->backButtonPressed(); break;
        case 47: _t->forwardButtonPressed(); break;
        case 48: _t->zoomOutButtonPressed(); break;
        case 49: _t->zoomInButtonPressed(); break;
        case 50: _t->zoomIndexChanged((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 51: _t->zoomEditingFinished(); break;
        case 52: _t->fitWidthButtonPressed(); break;
        case 53: _t->fitPageButtonPressed(); break;
        case 54: _t->selectModeButtonPressed(); break;
        case 55: _t->statusIndicatorPressed(); break;
        case 56: _t->findTextChanged(); break;
        case 57: _t->findNextButtonPressed(); break;
        case 58: _t->findPrevButtonPressed(); break;
        case 59: _t->newTabButtonPressed(); break;
        case 60: _t->switchTab((*reinterpret_cast< QListWidgetItem*(*)>(_a[1])),(*reinterpret_cast< QListWidgetItem*(*)>(_a[2]))); break;
        case 61: _t->infoComboBoxChanged((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 62: _t->outlineItemClicked((*reinterpret_cast< const QModelIndex(*)>(_a[1]))); break;
        case 63: _t->layerItemClicked((*reinterpret_cast< const QModelIndex(*)>(_a[1]))); break;
        case 64: _t->attachmentSaveClicked((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 65: _t->clearFindError(); break;
        default: ;
        }
    }
}

const QMetaObject XpdfViewer::staticMetaObject = {
    { &QMainWindow::staticMetaObject, qt_meta_stringdata_XpdfViewer.data,
      qt_meta_data_XpdfViewer,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *XpdfViewer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *XpdfViewer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_XpdfViewer.stringdata0))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int XpdfViewer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 66)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 66;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 66)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 66;
    }
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
